#!/bin/bash

# Original by black.dragon74

# Declare variables to be used in this script
homeDir="$(echo $HOME)"
scriptDir="$homeDir/Library/debugNk"
pledit=/usr/libexec/PlistBuddy
maskedVal="XX-MASKED-XX"
checkForConnAhead=0
randomNumber=$(echo $(( ( RANDOM )  + 12 )))
outDir="$homeDir/Desktop/$randomNumber"
panicDir="$outDir/panic_logs"
zipFileName=debug_$randomNumber.zip
efiloc="null"
sysPfl=/usr/sbin/system_profiler
hostName=$(hostname | sed 's/.local//g')
genSysDump="no"
fixDupes="no"
hasChamel=""
hasClover=""
epochFor1Day=86400
veriStat=""

function dumpKernelLog(){
	bt=$(sysctl -n kern.boottime | sed 's/^.*} //')

	bTm=$(echo "$bt" | awk '{print $2}')
	bTd=$(echo "$bt" | awk '{print $3}')
	bTt=$(echo "$bt" | awk '{print $4}')
	bTy=$(echo "$bt" | awk '{print $5}')

	bTm=$(awk -v "month=$bTm" 'BEGIN {months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec"; print (index(months, month) + 3) / 4}')
	bTm=$(printf %02d $bTm)

	ep=$(/bin/date -jf '%H:%M:%S' $bTt '+%s')

	cs=$((ep - 60 ))

	bTt=$(/bin/date -r $cs '+%H:%M:%S')

	stopTime=$(log show --debug --info --start "$bTy-$bTm-$bTd $bTt" | grep loginwindow | head -1)
	stopTime="${stopTime%      *}"

	echo "Extract boot log from $bTy-$bTm-$bTd $bTt"

	log show --debug --info --start "$bTy-$bTm-$bTd $bTt" | grep -E 'kernel:|loginwindow:' | sed -n -e "/kernel: PMAP: PCID enabled/,/$stopTime/ p"
}

function rebuildCaches(){
	swver=$(sw_vers -productVersion | sed 's/\.//g' | colrm 5)

	if [[ $swver -ge 1011 ]]; then
		# Using "kextcache system-caches" fixes redunancy but can't be used as it doesn't output anything when no invalid signatured kexts are found
		# Fall back to "sudo kextcache -i /" until a better approach is found (or Apple fixes it :P)
		sudo kextcache -i /
	else
		sudo touch /System/Library/Extensions && sudo kextcache -u / #Good old method
	fi

	# Fix redundancy on macOS High Sierra and above
	if [[ $swver -ge 1013 ]]; then
		fixDupes="yes"
	fi
}

function fixDupKextLog(){
	# Must be run after the kextcache log is generated
	cd "$outDir"
	mv kextcache_log.txt kextcache_log_temp.txt
	# Using sed as i love it, this can also be achieved using awk like: awk '{k=$8; $8=""} !a[$0]++{$8=k; print}' kextcache_log_temp.txt >>kextcache_log.txt
	cat kextcache_log_temp.txt | sed -E -e 's/OSKext 0x[^[:space:]]+ /OSKext CONST /g' | sort | uniq | tail -r >>kextcache_log.txt
	rm kextcache_log_temp.txt
}

function dumpKextstat(){

	echo -e "Report generated using gen_debug ver$scriptVersion\n"
	echo "ACPIPLAT LOG :-"
	kextstat|grep -y acpiplat
	echo "END ACPIPLAT LOG."
	echo -e " "
	echo -e " "


	echo -e "APPLEINTELCPU LOG:-"
	kextstat|grep -y appleintelcpu
	echo -e "END APPLEINTELCPU LOG."
	echo -e " "
	echo -e " "


	echo -e "APPLE LPC LOG:-"
	kextstat|grep -y applelpc
	echo -e "END APPLE LPC LOG."
	echo -e " "
	echo -e " "


	echo -e "APPLE HDA LOG:-"
	kextstat|grep -y applehda
	echo -e "END APPLE HDA LOG."
	echo -e " "
	echo -e " "


	echo -e "LS FOR APPLEHDA :-"
	ls -l /System/Library/Extensions/AppleHDA.kext/Contents/Resources/*.zml*
	echo -e "END LS FOR APPLEHDA."
	echo -e " "
	echo -e " "


	echo -e "PMSET DUMP :-"
	pmset -g 
	echo -e "END DUMP FOR PMSET"
	echo -e " "
	echo -e " "


	echo -e "PMSET ASSERTIONS DUMP :-"
	pmset -g assertions
	echo -e "END DUMP FOR PMSET ASSERTIONS."
	echo -e " "
	echo -e " "


	echo -e "DUMP FOR TRIM STATUS :-"
	system_profiler SPSerialATADataType|grep TRIM
	echo -e "END DUMP FOR TRIM STATUS."
	echo -e " "
	echo -e " "
}

function dumpAllKextstat(){
	kextstat
}

# Add function to dump system information, requested by Jake
# Make it optional though as a full system report might take 3+ minutes on slow machines
# If user includes a -sysprofile arg then only generate a system report
function genSystemRep(){
	# Generate report in .spx format so that it is easier to debug.
	# If user wishes so, he can generate a report in txt format.
	# To generate a report in .txt format you can use gen_debug -sysprofile txt
	if [[ ! -z $1 ]]; then
		# Check arg
		if [[ "$1" == "txt" ]]; then
			# Generate report in .txt format
			echo "Generating report in txt format as requested."
			$sysPfl > "$outDir/SysDump-$hostName.txt" 2>/dev/null
		else
			echo -e "Ignored invalid arg: $1\nGenerating report in spx format."
			# Generate report in spx format
			$sysPfl -xml > "$outDir/SysDump-$hostName.spx" 2>/dev/null
		fi
	else
		# Generate report in spx format
		$sysPfl -xml > "$outDir/SysDump-$hostName.spx" 2>/dev/null
	fi
}

function dumpBootLog(){
	if [[ -z $(which bdmesg) ]]; then
		echo "BDMESG not found. Unable to dump boot log."
	else
		bdmesg > "$outDir/bootlog.txt" 2>/dev/null
	fi
}

function dumpNVRAM(){
	nvram -x -p
}

function dumpPanicLog(){
	# If panic logs exist, dump them
	if [[ ! -z $(ls /Library/Logs/DiagnosticReports | grep "panic" | grep -i "kernel") ]];
	then
		echo "Dumping kernel panic logs..." 
		mkdir $panicDir
		for f in $(ls /Library/Logs/DiagnosticReports | grep "panic" | grep -i "kernel"); do
			cp /Library/Logs/DiagnosticReports/$f $panicDir
		done
	fi
}

function verifyModDate(){
	# Get current date
	dateNow="$(date +%m\ %d\ %Y | awk '{ for (i=NF; i>1; i--) printf("%s ",$i); print $1; }' | sed 's/\ /-/g')"

	# Get file's modification date
	dateMod="$(GetFileInfo -m "$1" | awk '{print $1}' | sed 's/\//\ /g' | awk '{ for (i=NF; i>1; i--) printf("%s ",$i); print $1; }' | sed 's/\ /-/g')"

	# Convert current date to epoch
	dateNowEpoch="$(date -j -f "%Y-%d-%m" "$dateNow" +%s)"

	# Convert modification date to epoch
	dateModEpoch="$(date -j -f "%Y-%d-%m" "$dateMod" +%s)"

	# Calculate Difference
	dateDiff="$(($dateNowEpoch - $dateModEpoch))"

	# Now we compare
	if [[ "$dateNowEpoch" == "$dateModEpoch" || $dateDiff -le $epochFor1Day ]]; then
		veriStat="true"
	elif [[ $dateDiff -gt $epochFor1Day ]]; then
		veriStat="false"
	fi
}

# Restrict execution to Darwin
if [[ "$(sw_vers -productName 2>/dev/null | awk '{print $1}')" != "Mac" ]]; then
	echo "This script is meant for Apple macOS/OSX only." && exit
fi

# Tell them noobs to hold their horses for a while
clear
echo "Please don't touch your computer until the program is finished."
read -p "Do we agree? [Yy/Nn]: " contResp
case $contResp in
	[yY]* )
		# Continue
		echo "Cool!" && sleep 0.4
		;;
	[nN]* )
		# Exit
		echo "Alrighty! See ya sometime soon."
		exit 1
		;;	
	* )
		# Exit
		echo "Invalid option selected. Aborting..."
		exit 1
		;;
esac

# Check if script directory exists else create it
if [[ -d $scriptDir ]];
	then
	echo -e "Found script data directory at $scriptDir"
else
	echo -e "Script data directory not present, creating it."
	mkdir -p "$scriptDir"
fi

# Start dumping the data, start by creating dirs
if [[ -e "$outDir" ]];
	then
	rm -rf "$outDir"
else
	mkdir -p "$outDir"
fi

# Now we listen for interrupts and then cleanup in case we need to exit prematurely.
trap '{ 

read -p "Hey, you pressed Ctrl+C. Quit?[yY/nN]: " quitResp ;
case $quitResp in
	[yY]* )
		echo "Umm.. Okay. Cleaning up and exiting, NOW!"
		rm -rf $outDir
		exit 1
		;;
	* )
		# Continue
		echo "Ah! Okay. Continuing then.." 
		;;
esac

}' INT

# Change active directory to $outDir
cd "$outDir"
echo -e "Data will be dumped at $outDir"

# Request root access
isRoot=$(sudo id -u 2>/dev/null)
if [[ $isRoot != 0 ]]; then
	echo "Root access failed. Aborting...."
	rm -rf $outDir
	exit -1
fi

# Dumping system logs
echo -e "Dumping System log."
cp /var/log/system.log .

# Dump kernel panics (if exists)
dumpPanicLog

# Dumping kernel log
echo -e "Dumping kernel log."
dumpKernelLog &> kernel_log.txt

# Dumping kextstat
echo -e "Dumping kextstat."
touch kextstat_log.txt
dumpKextstat &>kextstat_log.txt
dumpAllKextstat &>kextstat_all_log.txt

# Dump kextcache
echo -e "Dumping kextcache"
touch kextcache_log.txt
rebuildCaches &>kextcache_log.txt

# If needed, remove duplicate entries
if [[ $fixDupes == "yes" ]]; then
fixDupKextLog
fi

# Dump System Profile if user has asked so
if [[ "$genSysDump" == "yes" ]]; then
	echo "Generating system info, this may take a while."
	genSystemRep $2
else
	echo -e "System dump not requested.\nYou may use gen_debug -sysprofile to generate system dump."
	echo -e "For output in TXT format use: \"gen_debug -sysprofile txt\". Default format is SPX"
fi

# Dump boot log
echo "Dumping Boot log"
dumpBootLog

# Dump NVRAM
echo "Dumping NVRAM values..."
dumpNVRAM > nvram.plist

# Zip all the files
cd "$outDir"
echo -e "Zipping all the files"
zip -r $zipFileName * &>/dev/null
echo -e "Zipped files at: $outDir/debug_$randomNumber.zip"

# Remove unzipped files
cd "$outDir"
shopt -s extglob
rm -rf -- !(debug_*)

# Ask to open the out directory.
read -p "Dump complete. Open $outDir?(Yy/Nn) " readOut
case $readOut in
	[yY]|[yY][eE][sS] )
		open "$outDir"
		;;
	[nN]|[nN][oO] )
		echo -e "Okay. You can open it manually."
		;;	
	* )
		echo -e "Invalid option selected. Open manually."
		;;	
esac

# Say Thank You!
echo -e "Thank You! Hope your problem gets sorted out soon."
exit